from typing import Optional, Union, List, Dict

import ROOT

from quickstats.components import AnalysisBase
from quickstats.maths.numerics import is_float
from quickstats.utils.common_utils import format_delimiter_enclosed_text

class AsimovHandler(AnalysisBase):
    _DEFAULTS_ = {
        "raw": "raw",
        "fit": "fit",
        "reset": "reset",
        "gen_asimov": "genasimov",
        "float": "float",
        "fix_syst": "fixsyst",
        "fix_all": "fixall",
        "match_glob": "matchglob",
        "save_snapshot": "savesnapshot"
    }
    def __init__(self, ws:ROOT.RooWorkspace, data_name:str,
                 range_name:Optional[str]=None,
                 minimizer_config:Optional[Dict]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(ws, poi_name=[], data_name=data_name,
                         config=minimizer_config, verbosity=verbosity)
        self.range_name = range_name
        self.initialize()
    
    def initialize(self):
        self.core_variables = self.get_variables("core")
        self.save_snapshot(self._DEFAULTS_['raw'], self.core_variables)
        self.saved_snapshots = [self._DEFAULTS_['raw']]
        
    def parse_setup(self, setup_str:str):
        setup_str = setup_str.replace(" ", "")
        fix_setup = []
        profile_setup = []
        invalid_setup = []
        undefined_setup = []
        tokens = setup_str.split(",")
        for token in tokens:
            subtokens = token.split("=")
            if len(subtokens) != 2:
                invalid_setup.append(token)
                continue
            name = subtokens[0]
            expr = subtokens[1]
            if not self.model.workspace.var(name):
                undefined_setup.append(name)
                continue
            expr_tokens = expr.split("_")
            if len(expr_tokens) not in [1, 3]:
                invalid_setup.append(token)
                continue
            if len(expr_tokens) == 1:
                if not is_float(expr_tokens[0]):
                    invalid_setup.append(token)
                    continue
                else:
                    fix_setup.append(token)
            elif len(expr_tokens) == 3:
                if not all(is_float(t) for t in expr_tokens):
                    invalid_setup.append(token)
                    continue
                else:
                    profile_setup.append(token)
        passed = (len(invalid_setup) == 0) and (len(undefined_setup) == 0)
        if len(invalid_setup) > 0:
            invalid_setup_str = ", ".join([f"`{s}`" for s in invalid_setup])
            self.stdout.error(f"ERROR: Invalid setup info {invalid_setup_str}.", "red")
        if len(undefined_setup) > 0:
            undefined_setup_str = ", ".join([f"`{s}`" for s in undefined_setup])
            self.stdout.error(f"ERROR: Setup info with undefined variable {undefined_setup_str}.", "red")
        if passed:
            if (not fix_setup) and (not profile_setup):
                self.stdout.info("REGTEST: There is no setup info provided. "
                                 "Proceed with current variable values and ranges.")
            setups = {
                "fix_param"    : ",".join(fix_setup),
                "profile_param": ",".join(profile_setup)
            }
            return setups
        return None
    
    def translate_xml_attributes(self, attributes:Dict):
        result = {
            'name': attributes['Name'],
            'setup': attributes.get("Setup", ""),
            'action': attributes.get("Action", ""),
            'snapshot_all': attributes.get("SnapshotAll", ""),
            'snapshot_glob': attributes.get("SnapshotGlob", ""),
            'snapshot_nuis': attributes.get("SnapshotNuis", ""),
            'snapshot_poi': attributes.get("SnapshotPOI", ""),
            'data': attributes.get("Data", ""),
            'algorithm': attributes.get("Algorithm", "roostats")
        }
        return result
    
    def profile_to_data(self, data_name:str=""):
        tmp_data = self.model.data
        switch_data = (data_name != "") and (data_name != self.model.data.GetName())
        if switch_data:
            new_data = self.model.workspace.data(data_name)
            if not new_data:
                raise RuntimeError(f"dataset `{new_data}` not found in the workspace")
            self.set_data(new_data)
        # use range name when fitting to observed data
        if self.range_name and (not switch_data):
            self.set_range_name(self.range_name)
        fit_status = self.minimizer.minimize()
        # restore default dataset
        if switch_data:
            self.set_data(tmp_data)
        self.unset_range_name()
        return fit_status
    
    def set_range_name(self, range_name:str):
        self.minimizer.configure_nll(range=range_name, split_range=True, update=True)
        
    def unset_range_name(self):
        self.minimizer.nll = None
        self.minimizer.nll_commands.pop("RangeWithName", None)
        self.minimizer.nll_commands.pop("SplitRange", None)
    
    def generate_single_asimov(self, attributes:Dict):
        orig_snapshot = self.core_variables.snapshot()
        config = self.translate_xml_attributes(attributes)
        asimov_name = config['name']
        title_str = format_delimiter_enclosed_text(f"Operation {asimov_name}", "+")
        self.stdout.info(title_str)
        setup_str = config['setup'].strip()
        if setup_str:
            setup_kwargs = self.parse_setup(setup_str)
            if not setup_kwargs:
                self.stdout.error(f"\tStop generating Asimov data `{asimov_name}`")
                return None
            else:
                self.setup_parameters(**setup_kwargs, update_snapshot=False)
        action_str = config['action'].strip()
        self.stdout.info(f"REGTEST: Action list ({action_str})")
        if not action_str:
            return None
        action_list = action_str.split(":")
        fixed_variables = []
        for action in action_list:
            status = -1
            if action == self._DEFAULTS_['fit']:
                data_name = config['data']
                status = self.profile_to_data(data_name)
                if status not in [0, 1]:
                    raise RuntimeError("fit not converging properly")
            # reset to initial parameter values
            elif action == self._DEFAULTS_['raw']:
                self.load_snapshot(self._DEFAULTS_['raw'])
            # reset to parameter values at beginning of this round
            elif action == self._DEFAULTS_['reset']:
                self.core_variables.__assign__(orig_snapshot)
            # float fixed nuisance parameters
            elif action == self._DEFAULTS_['float']:
                for varname in fixed_variables:
                    if self.model.pois.find(varname):
                        continue
                    else:
                        self.model.workspace.var(varname).setConstant(False)
            elif action == self._DEFAULTS_['gen_asimov']:
                self.stdout.info(f"REGTEST: Generating Asimov dataset `{asimov_name}`")
                algorithm = config['algorithm']
                if algorithm == "roostats":
                    generator = ROOT.RooStats.AsymptoticCalculator.GenerateAsimovData
                    asimov_data = generator(self.model.pdf, self.model.observables)
                    getattr(self.model.workspace, "import")(asimov_data, ROOT.RooFit.Rename(asimov_name))
                else:
                    asimov_data = self.model.generate_asimov(do_fit=False, modify_globs=False, do_import=True,
                                                             asimov_name=asimov_name, restore_status=0)
            elif action == self._DEFAULTS_['fix_syst']:
                constrained_np = self.model.get_constrained_nuisance_parameters()
                for np in constrained_np:
                    np.setConstant(True)
                fixed_variables += [np.GetName() for np in constrained_np]
            elif action == self._DEFAULTS_['fix_all']:
                all_np = self.model.nuisance_parameters
                for np in all_np:
                    np.setConstant(True)
                fixed_variables += [np.GetName() for np in all_np]
            elif action == self._DEFAULTS_['match_glob']:
                if len(self.model.global_observables) > 0:
                    self.model.match_globs()
            elif action == self._DEFAULTS_['save_snapshot']:
                param_str_map = {
                    'all': 'parameters of interest, nuisance parameters, and global observables',
                    'nuis': 'nuisance parameters',
                    'glob': 'global observables',
                    'poi': 'parameters of interest'
                }
                param_var_map = {
                    'all': self.core_variables,
                    'nuis': self.model.nuisance_parameters,
                    'glob': self.model.global_observables,
                    'poi': self.model.pois,
                }                
                for param, param_str in param_str_map.items():
                    snapshot_name = config[f"snapshot_{param}"]
                    if snapshot_name != "":
                        self.stdout.info(f"REGTEST: Saving snapshot {snapshot_name} for current {param_str}")
                        self.save_snapshot(snapshot_name, param_var_map[param])
                        self.saved_snapshots.append(snapshot_name)
            elif action in self.saved_snapshots:
                self.load_snapshot(action)
            else:
                raise RuntimeError(f"unknown action {action}")


import os
import click

@click.command(name='build_xml_ws')
@click.option('-i', '--filename', required=True, 
              help='Input xml file.')
@click.option('--binned/--unbinned', 'use_binned', default=False, show_default=True,
              help='Fit to binned data.')
@click.option('--data_storage_type', default="vector", show_default=True, 
              type=click.Choice(['vector', 'tree', 'composite']),
              help='Set RooAbsData StorageType. Available choices: "vector", "tree", "composite".')
@click.option('-d', '--basedir', default=None, 
              help='Base directory to which files in the xmls are referenced. '
                   'By default, the directory of the input xml file is used.')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--retry', type=int, default=0, show_default=True,
              help='Maximum number of retries upon a failed fit')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('-v', '--verbosity',  default="INFO", show_default=True,
              help='verbosity level ("DEBUG", "INFO", "WARNING", "ERROR")')
def build_xml_ws(**kwargs):
    """
    Build workspace from XML config files
    """
    _kwargs = {}
    for arg_name in ["filename", "use_binned", "data_storage_type",
                     "basedir", "verbosity"]:
        _kwargs[arg_name] = kwargs.pop(arg_name)
    _kwargs['minimizer_config'] = kwargs
    from quickstats.components.workspaces import XMLWSBuilder
    builder = XMLWSBuilder(**_kwargs)
    builder.generate_workspace()


import sys
from enum import Enum
from typing import Union
from functools import total_ordering

text_color_map = {
    'black': '\033[30m',
    'red': '\033[31m',
    'green': '\033[32m',
    'yellow': '\033[33m',
    'blue': '\033[34m',
    'magenta': '\033[35m',
    'cyan': '\033[36m',
    'white': '\033[37m',
    'bright black': '\033[30;1m',
    'bright red': '\033[31;1m',
    'bright green': '\033[32;1m',
    'bright yellow': '\033[33;1m',
    'bright blue': '\033[34;1m',
    'bright magenta': '\033[35;1m',
    'bright cyan': '\033[36;1m',
    'bright white': '\033[37;1m',    
    'darkred': '\033[91m',
    'reset': '\033[0m'
}

@total_ordering
class Verbosity(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0
    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        elif isinstance(other, int):
            return self.value < other
        elif isinstance(other, str):
            return self.value < getattr(self, other.upper()).value
        return NotImplemented

class VerbosePrint:
    @property
    def verbosity(self):
        return self._verbosity
    
    @verbosity.setter
    def verbosity(self, val):
        if isinstance(val, str):
            try:
                v = getattr(Verbosity, val.upper())
            except:
                raise ValueError("invalid verbosity level: {}".format(val))
            self._verbosity = v
        else:
            self._verbosity = val

    def __init__(self, verbosity:Union[int, Verbosity, str]=Verbosity.INFO):
        self.verbosity = verbosity
        
    def info(self, text:str, color=None):
        self.__call__(text, Verbosity.INFO, color=color)
        
    def warning(self, text:str, color=None):
        self.__call__(text, Verbosity.WARNING, color=color)
        
    def error(self, text:str, color=None):
        self.__call__(text, Verbosity.ERROR, color=color)
        
    def critical(self, text:str, color=None):
        self.__call__(text, Verbosity.CRITICAL, color=color)

    def debug(self, text:str, color=None):
        self.__call__(text, Verbosity.DEBUG, color=color)
        
    def __call__(self, text:str, verbosity:int=Verbosity.INFO,
                 color=None):
        if color:
            text = f"{text_color_map[color]}{text}{text_color_map['reset']}"
        if verbosity >= self.verbosity:
            sys.stdout.write(text + "\n")


import os
import yaml
from typing import Optional, Dict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

STYLES = {
    "default": {
        'cmap': 'seismic',
        'value_color': 'black'
    },
    "viridis":{
        'cmap': 'viridis',
        'value_color': 'black'
    }
}

class CorrelationPlot():
    
    def __init__(self, data:pd.DataFrame):
        
        self.data = data
        
    @staticmethod
    def parse_style(style:Optional[str]=None):
        if not style:
            return STYLES["default"]
        elif isinstance(style, dict):
            return style
        if os.path.exists(style):
            return yaml.safe_load(open(style))
        _style = STYLES.get(style.lower(), None)
        if not _style:
            raise ValueError(f'unknown style: {style}')
        return _style
        
    def draw(self, cmap:str='seismic', xlabel_rotation:float=90, ylabel_rotation:float=0, label_size:float=25,
              figscale:int=1, show_values:bool=True, value_color:str="black", value_size=16, value_precision:int=2,
              gridline:Optional[str]="--", gridcolor:str="black"):
        plt.clf()
        size = len(self.data)
        figsize = size * figscale
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(figsize, figsize), facecolor= "#FFFFFF", dpi=72)
        ax.matshow(self.data, cmap=cmap, vmin=-1, vmax=1)
        ticks = np.arange(0, len(self.data), 1)
        ax.set_xticks(ticks)
        ax.set_yticks(ticks)
        ax.set_xticklabels(self.data.columns, rotation=xlabel_rotation, fontsize=label_size)
        ax.set_yticklabels(self.data.index, rotation=ylabel_rotation, fontsize=label_size)
        ax.tick_params(axis="both", which="both", length=0)
        if show_values:
            fmt_str = f'{{:0.{value_precision}f}}'
            for (i, j), z in np.ndenumerate(self.data.values):
                ax.text(j, i, fmt_str.format(z), ha='center', va='center', color=value_color, fontsize=value_size)
        if gridline:
            for i in range(len(ticks) - 1):
                ax.axhline(i + 0.5, linestyle=gridline, color=gridcolor)
                ax.axvline(i + 0.5, linestyle=gridline, color=gridcolor)
        return ax
    
    def draw_style(self, style:Optional[Dict]=None):
        style = self.parse_style(style)
        return self.draw(**style)



import os
import json
import click

@click.command(name='run_pulls')
@click.option('-i', '--input_file', required=True, 
              help='Path to the input workspace file')
@click.option('-x', '--poi', required=True,
              help='POIs to measure')
@click.option('-o', '--outdir', default="pulls", show_default=True,
              help='Output directory')
@click.option('-w', '--workspace', default=None, 
              help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', default=None, 
              help='Name of model config. Auto-detect by default.')
@click.option('-d', '--data', default='combData', show_default=True,
              help='Name of dataset')
@click.option('-p', '--parameter', default='', show_default=True,
              help='Nuisance parameter(s) to run pulls on.'+\
                   'Multiple parameters are separated by commas.'+\
                   'Wildcards are accepted. All NPs are run by default.')
@click.option('-r', '--profile', default=None, show_default=True,
              help='Parameters to profile')
@click.option('-f', '--fix', default=None, show_default=True,
              help='Parameters to fix')
@click.option('-s', '--snapshot', default="nominalNuis", show_default=True,
              help='Name of initial snapshot')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('--binned/--unbinned', default=True, show_default=True,
              help='Binned likelihood')
@click.option('-q', '--precision', type=float, default=0.001, show_default=True,
              help='Precision for scan')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--retry', type=int, default=0, show_default=True,
              help='Maximum number of retries upon a failed fit')
@click.option('-l', '--log_level', default="INFO", show_default=True,
              help='Log level')
@click.option('--print_level', type=int, default=1, show_default=True,
              help='Minimizer print level')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default strategy')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-multi', default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--batch_mode/--no-batch', default=False, show_default=True,
              help='Batch mode when evaluating likelihood')
@click.option('--int_bin_precision', type=float, default=-1., show_default=True,
              help='Integrate the PDF over the bins instead of using the probability '
                   'density at the bin centre')
@click.option('--parallel', type=int, default=-1, show_default=True,
              help='\b\n Parallelize job across the N workers.'
                   '\b\n Case  0: Jobs are run sequentially (for debugging).'
                   '\b\n Case -1: Jobs are run across N_CPU workers.')
@click.option('--cache/--no-cache', default=True, show_default=True,
              help='Cache existing result')
@click.option('--exclude', default="", show_default=True,
              help='Exclude NPs (wildcard is accepted)')
def run_pulls(**kwargs):
    """
    Tool for computing NP pulls and impacts
    """
    from quickstats.components import NuisanceParameterPull
    NuisanceParameterPull().run_pulls(**kwargs)
    
@click.command(name='plot_pulls')
@click.option('-i', '--inputdir', required=True, help='Path to directory containing pull results')
@click.option('-p', '--poi', default=None, help='Parameter of interest for plotting impact')
@click.option('-n', '--n_rank', type=int, default=None, help='Total number of NP to rank')
@click.option('-m', '--rank_per_plot', type=int, default=20, show_default=True,
              help='Number of NP to show in a single plot')
@click.option('--ranking/--no_ranking', default=True, show_default=True,
              help='Rank NP by impact')
@click.option('--threshold', type=float, default=0., show_default=True,
              help='Filter NP by postfit impact threshold')
@click.option('--show_sigma/--hide_sigma', default=True, show_default=True,
              help='Show one standard deviation pull')
@click.option('--show_prefit/--hide_prefit', default=True, show_default=True,
              help='Show prefit impact')
@click.option('--show_postfit/--hide_postfit', default=True, show_default=True,
              help='Show postfit impact')
@click.option('--sigma_bands/--no_sigma_bands', default=False, show_default=True,
              help='Draw +-1, +-2 sigma bands')
@click.option('--sigma_lines/--no_sigma_lines', default=True, show_default=True,
              help='Draw +-1 sigma lines')
@click.option('--ranking_label/--no_ranking_label', default=True, show_default=True,
              help='Show ranking labels')
@click.option('--shade/--no_shade', default=True, show_default=True,
              help='Draw shade')
@click.option('--correlation/--no_correlation', default=True, show_default=True,
              help='Show correlation impact')
@click.option('--onesided/--overlap', default=True, show_default=True,
              help='Show onesided impact')
@click.option('--relative/--absolute', default=False, show_default=True,
              help='Show relative variation')
@click.option('--theta_max', type=float, default=2, show_default=True,
              help='Pull range')
@click.option('-y', '--padding', type=int, default=7, show_default=True,
              help='Padding below plot for texts and legends. NP column height is 1 unit.')
@click.option('-h', '--height', type=float, default=1.0, show_default=True,
              help='NP column height')
@click.option('-s', '--spacing', type=float, default=0., show_default=True,
              help='Spacing between impact box')
@click.option('-d', '--display_poi', default=r"$\mu$", show_default=True,
              help='POI name to be shown in the plot')
@click.option('-t', '--extra_text', default=None, help='Extra texts below the ATLAS label. '+\
                                                       'Use "//" as newline delimiter')
@click.option('--elumi_label/--no_elumi_label', default=True, show_default=True,
              help='Show energy and luminosity labels')
@click.option('--ranking_label/--no_ranking_label', default=True, show_default=True,
              help='Show ranking label')
@click.option('--energy', default="13 TeV", show_default=True, 
              help='Beam energy')
@click.option('--lumi', default="139 fb$^{-1}$", show_default=True, 
              help='Luminosity')
@click.option('--status', default="int", show_default=True, 
              help='\b\n Analysis status. Choose from'
                   '\b\n            int : Internal'
                   '\b\n            wip : Work in Progress'
                   '\b\n         prelim : Preliminary'
                   '\b\n          final : *no status label*'
                   '\b\n *custom input* : *custom input*')
@click.option('--combine_pdf/--split_pdf', default=True, show_default=True,
              help='Combine all ranking plots into a single pdf')
@click.option('--outdir', default='ranking_plots', show_default=True,
              help='Output directory')
@click.option('-o', '--outname', default='ranking', show_default=True,
              help='Output file name prefix')
@click.option('--style', default='default', show_default=True,
              help='Plotting style. Built-in styles are "default" and "trex".'+\
                   'Specify path to yaml file to set custom plotting style.')
@click.option('--fix_axis_scale/--free_axis_scale', default=True, show_default=True,
              help='Fix the axis scale across all ranking plots')
def plot_pulls(**kwargs):
    """
    Tool for plotting NP pulls and impact rankings
    """    
    from quickstats.plots.np_ranking_plot import NPRankingPlot
    inputdir, poi = kwargs.pop('inputdir'), kwargs.pop('poi')
    ranking_plot = NPRankingPlot(inputdir, poi)
    ranking_plot.plot(**kwargs)

@click.command(name='cls_limit')
@click.option('-i', '--input_file', 'filename', required=True, 
              help='Path to the input workspace file')
@click.option('-p', '--poi', 'poi_name', default=None, show_default=True,
              help='POI to scan. If not specified, the first POI from the workspace is used.')
@click.option('-d', '--data', 'data_name', default='combData', show_default=True,
              help='Name of dataset')
@click.option('-o', '--outname', default='limits.json', show_default=True,
              help='Name of output')
@click.option('--blind/--unblind', 'do_blind', default=True, show_default=True,
              help='Blind/unblind analysis')
@click.option('--CL', 'CL', type=float, default=0.95, show_default=True,
              help='CL value to use')
@click.option('--precision', default=0.005, show_default=True,
              help='precision in mu that defines iterative cutoff')
@click.option('--do_tilde/--no_tilde', default=True, show_default=True,
              help='bound mu at zero if true and do the \tilde{q}_{mu} asymptotics')
@click.option('--predictive_fit/--no_predictive_fit', default=True, show_default=True,
              help='extrapolate best fit nuisance parameters based on previous fit results')
@click.option('--do_better_bands/--skip_better_bands', default=True, show_default=True,
              help='evaluate asymptotic CLs limit for various sigma bands')
@click.option('--better_negative_bands/--skip_better_negative_bands', default=False, show_default=True,
              help='evaluate asymptotic CLs limit for negative sigma bands')
@click.option('--binned/--unbinned', 'binned_likelihood', default=True, show_default=True,
              help='Binned likelihood')
@click.option('--save_summary/--skip_summary', default=True, show_default=True,
              help='Save summary information')
@click.option('-f', '--fix', 'fix_param', default="", show_default=True,
              help='Parameters to fix')
@click.option('-r', '--profile', 'profile_param', default="", show_default=True,
              help='Parameters to profile')
@click.option('-w', '--workspace', 'ws_name', default=None, show_default=True,
              help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, show_default=True,
              help='Name of model config. Auto-detect by default.')
@click.option('-s', '--snapshot', 'snapshot_name', default=None, show_default=True,
              help='Name of initial snapshot')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--retry', type=int, default=0, show_default=True,
              help='Maximum number of retries upon a failed fit')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('--timer/--no_timer', default=False, show_default=True,
              help='Enable minimizer timer')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--batch_mode/--no-batch', default=False, show_default=True,
              help='Batch mode when evaluating likelihood')
@click.option('--int_bin_precision', type=float, default=-1., show_default=True,
              help='Integrate the PDF over the bins instead of using the probability '
                   'density at the bin centre')
@click.option('--constrain/--no-constrain', 'constrain_nuis', default=True, show_default=True,
              help='Use constrained NLL')
@click.option('-v', '--verbosity', default='INFO', show_default=True,
              type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
              help='Verbosity level.')
def cls_limit(**kwargs):
    """
    Tool for evaluating Asymptotic CLs limit
    """
    from quickstats.components import AsymptoticCLs
    outname = kwargs.pop('outname')
    save_summary = kwargs.pop('save_summary')
    asymptotic_cls = AsymptoticCLs(**kwargs)
    asymptotic_cls.evaluate_limits()
    asymptotic_cls.save(outname, summary=save_summary)

@click.command(name='compile')
@click.option('-m', '--macros', default=None, show_default=True,
              help='Macros to compile (separated by commas). By default all macros are compiled.')
def compile_macros(macros):
    """
    Compile ROOT macros
    """
    import quickstats
    quickstats.compile_macros(macros)

@click.command(name='harmonize_np')
@click.argument('ws_files', nargs=-1)
@click.option('-r', '--reference', required=True, help='Path to reference json file containing renaming scheme')
@click.option('-i', '--input_config_path', default=None, show_default=True,
              help='Path to json file containing input workspace paths')
@click.option('-b', '--base_path', default='./', show_default=True,
              help='Base path for input config')
@click.option('-o', '--outfile', default='renamed_np.json', show_default=True,
              help='Output filename')
def harmonize_np(ws_files, reference, input_config_path, base_path, outfile):
    """
    Harmonize NP names across different workspaces
    """
    from quickstats.components import NuisanceParameterHarmonizer
    harmonizer = NuisanceParameterHarmonizer(reference)
    if (len(ws_files) > 0) and input_config_path is not None:
        raise RuntimeError('either workspace paths or json file containing workspace paths should be given')
    if len(ws_files) > 0:
        harmonizer.harmonize(ws_files, outfile=outfile)
    elif (input_config_path is not None):
        harmonizer.harmonize_multi_input(input_config_path, base_path, outfile=outfile)
        
        
@click.command(name='generate_asimov')
@click.option('-i', '--input_file', required=True, 
              help='Path to the input workspace file.')
@click.option('-o', '--output_file', required=True, 
              help='Name of the output workspace containing the '
                   'generated asimov dataset.')
@click.option('-p', '--poi', required=True, 
              help='Name of the parameter of interest (POI).')
@click.option('--poi_val', type=float, default=None, show_default=True,
              help='Generate asimov data with POI set at the specified value. '
                   'If None, POI will be kept at the post-fit value if a fitting '
                   'is performed or the pre-fit value if no fitting is performed.')
@click.option('--poi_profile', type=float, default=None, show_default=True,
              help='Perform nuisance parameter profiling with POI set at the specified value. '
                   'This option is only effective if do_fit is set to True. If None, POI is '
                   'set floating (i.e. unconditional maximum likelihood estimate).')
@click.option('--modify-globs/--keep-globs', default=True, show_default=True,
              help='Match the values of nuisance parameters and the corresponding global '
                   'observables when generating the asimov data. This is important for making '
                   'sure the asimov data has the (conditional) minimal NLL.')
@click.option('--do-fit/--no-fit', default=True, show_default=True,
              help='Perform nuisance parameter profiling with a fit to the given dataset.')
@click.option('--asimov_name', default="asimovData_{mu}", show_default=True,
              help='Name of the generated asimov dataset.')
@click.option('-d', '--data', default='combData', show_default=True,
              help='Name of the dataset used in NP profiling.')
@click.option('--constraint_option', default=0, show_default=True,
              help='\b\n Customize the target of nuisance paramaters involved in the profiling.'
                   '\b\n Case 0: All nuisance parameters are allowed to float;'
                   '\b\n Case 1: Constrained nuisance parameters are fixed to 0.'
                   '\b\n         Unconstrained nuisrance parameters are allowed to float.')
@click.option('-c', '--configuration', default=None,
              help='Path to the json configuration file containing'
                   ' the minimizer options for NP profiling.')
def generate_asimov(**kwargs):
    """
    Generate Asimov dataset
    """
    input_filename = kwargs.pop("input_file")
    output_filename = kwargs.pop("output_file")
    if os.path.abspath(input_filename) == os.path.abspath(output_filename):
        raise ValueError("output workspace file name cannnot be the same as the input workspace file name")
    from quickstats.components import AnalysisObject
    config_file = kwargs.pop("configuration")
    if config_file is not None:
        config = json.load(open(config_file, 'r'))
    else:
        config = {}
    data_name = kwargs.pop("data")
    kwargs['poi_name'] = kwargs.pop("poi")
    task = AnalysisObject(filename=input_filename, data_name=data_name, **config)
    task.model.generate_asimov(**kwargs)
    task.model.save(output_filename)


@click.command(name='generate_standard_asimov')
@click.option('-i', '--input_file', 'filename', required=True, 
              help='Path to the input workspace file.')
@click.option('-o', '--output_file', 'outname', required=True, 
              help='Name of the output workspace containing the '
                   'generated asimov dataset.')
@click.option('-d', '--data', 'data_name', default='combData', show_default=True,
              help='Name of the dataset used in NP profiling.')
@click.option('-p', '--poi', 'poi_name', required=True, 
              help='Name of the parameter of interest (POI).')
@click.option('-s', '--poi_scale', type=float, default=1.0, show_default=True,
              help='Scale factor applied to the poi value')
@click.option('-f', '--fix', 'fix_param', default="", show_default=True,
              help='Parameters to fix')
@click.option('-r', '--profile', 'profile_param', default="", show_default=True,
              help='Parameters to profile')
@click.option('-s', '--snapshot', 'snapshot_name', default=None, show_default=True,
              help='Name of initial snapshot')
@click.option('--asimov_names', default=None, show_default=True,
              help='Names of the output asimov datasets (separated by commas). If not specified, '
                   'a default name for the corresponding asimov type will be given.')
@click.option('--asimov_snapshots', default=None, show_default=True,
              help='Names of the output asimov snapshots (separated by commas). If not specified, '
                   'a default name for the corresponding asimov type will be given.')
@click.option('-t', '--asimov_types', default="0,1,2", show_default=True,
              help='\b\n Types of asimov dataset to generate separated by commas.'
                   '\b\n 0: fit with POI fixed to 0'
                   '\b\n 1: fit with POI fixed to 1'
                   '\b\n 2: fit with POI free and set POI to 1 after fit'
                   '\b\n 3: fit with POI and constrained NP fixed to 0'
                   '\b\n 4: fit with POI fixed to 1 and constrained NP fixed to 0'
                   '\b\n 5: fit with POI free and constrained NP fixed to 0 and set POI to 1 after fit'
                   '\b\n -1: nominal NP with POI set to 0'
                   '\b\n -2: nominal NP with POI set to 1')
def generate_standard_asimov(**kwargs):
    """
    Generate standard Asimov dataset
    """
    from quickstats.components import AsimovGenerator
    outname = kwargs.pop('outname')
    asimov_types = kwargs.pop('asimov_types')
    asimov_types = [int(t) for t in asimov_types.split(',')]
    fix_param = kwargs.pop('fix_param')
    profile_param = kwargs.pop('profile_param')
    snapshot_name = kwargs.pop('snapshot_name')
    poi_scale = kwargs.pop("poi_scale")
    asimov_names = kwargs.pop("asimov_names")
    asimov_snapshots = kwargs.pop("asimov_snapshots")
    config = {
        'fix_param': fix_param,
        'profile_param': profile_param,
        'snapshot_name': snapshot_name
    }
    if asimov_names is not None:
        asimov_names = asimov_names.split(",")
    if asimov_snapshots is not None:
        asimov_snapshots = asimov_snapshots.split(",")        
    generator = AsimovGenerator(**kwargs, config=config)
    generator.generate_standard_asimov(asimov_types, poi_scale=poi_scale,
                                       asimov_names=asimov_names,
                                       asimov_snapshots=asimov_snapshots)
    generator.save(outname)
    

@click.command(name='toy_significance')
@click.option('-i', '--input_file', 'filename', required=True, 
              help='Path to the input workspace file.')
@click.option('-o', '--output_file', 'outname', default="toy_study/results.json", 
              help='Name of the output file containing toy results.')
@click.option('-n', '--n_toys', type=int,
              help='Number of the toys to use.')
@click.option('-b', '--batchsize', type=int, default=100, show_default=True,
              help='Divide the task into batches each containing this number of toys. '
                   'Result from each batch is saved for caching and different batches '
                   'are run in parallel if needed')
@click.option('-s', '--seed', type=int, default=0,  show_default=True,
              help='Random seed used for generating toy datasets.')
@click.option('-p', '--poi', 'poi_name', default=None,
              help='Name of the parameter of interest (POI). If None, the first POI is used.')
@click.option('-v', '--poi_val', type=float, default=0,  show_default=True,
              help='POI value when generating the toy dataset.')
@click.option('--binned/--unbinned', default=True, show_default=True,
              help='Generate binned toy dataset.')
@click.option('--cache/--no-cache', default=True,  show_default=True,
              help='Cache existing batch results.')
@click.option('--fit_options', default=None, help='A json file specifying the fit options.')
@click.option('-v', '--verbosity', default='INFO', show_default=True,
              type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
              help='Verbosity level.')
@click.option('--parallel', type=int, default=-1, show_default=True,
              help='\b\n Parallelize job across the N workers.'
                   '\b\n Case  0: Jobs are run sequentially (for debugging).'
                   '\b\n Case -1: Jobs are run across N_CPU workers.')
def toy_significance(**kwargs):
    """
    Generate toys and evaluate significance
    """
    from quickstats.components import PValueToys
    n_toys = kwargs.pop("n_toys")
    batchsize = kwargs.pop("batchsize")
    seed = kwargs.pop("seed")
    cache = kwargs.pop("cache")
    outname = kwargs.pop("outname")
    parallel = kwargs.pop("parallel")
    pvalue_toys = PValueToys(**kwargs)
    pvalue_toys.get_toy_results(n_toys=n_toys, batchsize=batchsize, seed=seed,
                                cache=cache, save_as=outname, parallel=parallel)
    
    
    
@click.command(name='toy_limit')
@click.option('-i', '--input_file', 'filename', required=True, 
              help='Path to the input workspace file.')
@click.option('-d', '--data', 'data_name', default='combData', show_default=True,
              help='Name of the dataset used for computing observed limit.')
@click.option('-o', '--output_file', 'outname', 
              default="toy_study/toy_result_seed_{seed}_batch_{batch}.root",
              show_default=True,
              help='Name of the output file containing toy results.')
@click.option('--poi_max', type=float, default=None,
              help='Maximum range of POI.')
@click.option('--poi_min', type=float, default=None,
              help='Minimum range of POI.')
@click.option('--scan_max', type=float, default=None,
              help='Maximum scan value of POI.')
@click.option('--scan_min', type=float, default=None,
              help='Minimum scan value of POI.')
@click.option('--steps', type=int, default=10, show_default=True,
              help='Number of scan steps.')
@click.option('--mu_val', type=float, default=None,
              help='Value of POI for running a single point')
@click.option('-n', '--n_toys', type=int,
              help='Number of the toys to use.')
@click.option('-b', '--batchsize', type=int, default=50, show_default=True,
              help='Divide the task into batches each containing this number of toys. '
                   'Result from each batch is saved for caching and different batches '
                   'are run in parallel if needed')
@click.option('-s', '--seed', type=int, default=2021,  show_default=True,
              help='Random seed used for generating toy datasets.')
@click.option('-t', '--tolerance', type=float, default=1.,  show_default=True,
              help='Tolerance for minimization.')
@click.option('-p', '--poi', 'poi_name', default=None,
              help='Name of the parameter of interest (POI). If None, the first POI is used.')
@click.option('--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Use NLL offset.')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('-v', '--verbosity', default='INFO', show_default=True,
              type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"], case_sensitive=False),
              help='Verbosity level.')
@click.option('--snapshot', 'snapshot_name', default=None, help='Name of initial snapshot')
@click.option('--parallel', type=int, default=-1, show_default=True,
              help='\b\n Parallelize job across the N workers.'
                   '\b\n Case  0: Jobs are run sequentially (for debugging).'
                   '\b\n Case -1: Jobs are run across N_CPU workers.')
def toy_limit(**kwargs):
    """
    Generate toys and evaluate limits
    """
    from quickstats.components.toy_limit_calculator import evaluate_batched_toy_limits
    if not (((kwargs['scan_min'] is None) and (kwargs['scan_max'] is None) and (kwargs['mu_val'] is not None)) or \
           ((kwargs['scan_min'] is not None) and (kwargs['scan_max'] is not None) and (kwargs['mu_val'] is None))):
        raise ValueError("please provide either (scan_min, scan_max, steps) for running a scan or (mu_val)"
                         " for running a single point")        
    evaluate_batched_toy_limits(**kwargs)